'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var mimic = require('mimic-fn');
require('weakmap-polyfill');

function chain(firstIterable, secondIterable) {
  if (firstIterable === EMPTY_ITERABLE) {
    return secondIterable;
  } else if (secondIterable === EMPTY_ITERABLE) {
    return firstIterable;
  }
  return {
    [Symbol.iterator]() {
      let iter = firstIterable[Symbol.iterator]();
      let firstDone = false;
      return {
        next() {
          for (; ; ) {
            const { value, done } = iter.next();
            if (done) {
              if (!firstDone) {
                iter = secondIterable[Symbol.iterator]();
                firstDone = true;
              } else {
                return ITER_DONE_VALUE;
              }
            } else {
              return { value, done: false };
            }
          }
        }
      };
    }
  };
}
function append(iterable, value) {
  return {
    [Symbol.iterator]() {
      let state = 0;
      const iter = iterable[Symbol.iterator]();
      return {
        next() {
          if (state === 0) {
            const item = iter.next();
            if (item.done) {
              state = 1;
            } else {
              return item;
            }
          }
          if (state === 1) {
            state = 2;
            return { value, done: false };
          }
          return ITER_DONE_VALUE;
        }
      };
    }
  };
}
function map(fn, iterable) {
  return {
    [Symbol.iterator]() {
      const iter = iterable[Symbol.iterator]();
      return {
        next() {
          const { value, done } = iter.next();
          return done ? ITER_DONE_VALUE : { value: fn(value), done };
        }
      };
    }
  };
}
function flatMap(map2, iterable) {
  return {
    [Symbol.iterator]() {
      const rootIter = iterable[Symbol.iterator]();
      let lastIter = null;
      return {
        next() {
          for (; ; ) {
            if (lastIter == null) {
              const { value: nextIter, done } = rootIter.next();
              if (!done) {
                lastIter = map2(nextIter)[Symbol.iterator]();
              } else {
                return ITER_DONE_VALUE;
              }
            }
            if (lastIter) {
              const item = lastIter.next();
              if (item.done) {
                lastIter = null;
              } else {
                return { value: item.value, done: false };
              }
            }
          }
        }
      };
    }
  };
}
function empty() {
  return EMPTY_ITERABLE;
}
function once(value) {
  return {
    [Symbol.iterator]() {
      let taken = false;
      return {
        next() {
          if (!taken) {
            taken = true;
            return { value, done: false };
          } else {
            return ITER_DONE_VALUE;
          }
        }
      };
    }
  };
}
function forEach(fn, iterable) {
  if (iterable === EMPTY_ITERABLE)
    return;
  const iter = iterable[Symbol.iterator]();
  for (let item = iter.next(); !item.done; fn(item.value), item = iter.next())
    ;
}
function double(first, second) {
  return {
    [Symbol.iterator]() {
      let state = 0;
      return {
        next() {
          if (state === 0) {
            state = 1;
            return { value: first, done: false };
          }
          if (state === 1) {
            state = 2;
            return { value: second, done: false };
          }
          return ITER_DONE_VALUE;
        }
      };
    }
  };
}
const ITER_DONE_VALUE = { value: void 0, done: true };
const EMPTY_ITERATOR = {
  next() {
    return ITER_DONE_VALUE;
  }
};
const EMPTY_ITERABLE = {
  [Symbol.iterator]() {
    return EMPTY_ITERATOR;
  }
};

const NO_VALUE = {};
const equals = (v1, v2) => v1 === v2 || v2 !== v2 && v1 !== v1;
function isPrimitiveValue(value) {
  if (value == null) {
    return true;
  } else {
    const type = typeof value;
    return type !== "object" && type !== "function";
  }
}

const noop = () => {
};
const mapImplemented = () => Boolean(Map && Map.prototype && typeof Map.prototype.entries === "function");
const EMPTY_ARRAY = [];
const EMPTY_OBJECT = {};
class HasLength {
  constructor() {
  }
  /**
   * Returns `true` if the underlying container is empty (has zero length).
   */
  isEmpty() {
    return this.len() === 0;
  }
}
class HasCapacity extends HasLength {
  /**
   * Creates a new container with supplied capacity.
   * In case capacity is equal to zero, an error will be thrown.
   * @param capacity
   */
  constructor(capacity) {
    super();
    if (capacity === 0) {
      throw new Error("Capacity can't be equal to zero");
    }
    this._capacity = capacity;
  }
  /**
   * Returns capacity of the underlying container.
   *
   *
   */
  capacity() {
    return this._capacity;
  }
  /**
   * Returns `true` if the length of the underlying container reached its capacity.
   *
   *
   */
  isFull() {
    return this.capacity() < this.len();
  }
  /**
   * Returns `true` if the length of the underlying container will reach (or already reached) its capacity.
   *
   *
   */
  willBeFull() {
    return this.capacity() < this.len() + 1;
  }
}
class ParentPath {
  constructor(parent, key) {
    this.parent = parent;
    this.key = key;
  }
  /// Enforces parent to drop the supplied child. Its value will be used only in case if key is absent.
  drop(child) {
    this.parent.drop(this.key === NO_VALUE ? child : this.key);
  }
}
const _Result = class {
  constructor(removed = EMPTY_ITERABLE, added = EMPTY_ITERABLE) {
    this.removed = removed;
    this.added = added;
  }
  /** Creates a result containing added entities */
  static added(added) {
    return new _Result(void 0, added);
  }
  /** Creates a result containing removed entities */
  static removed(removed) {
    return new _Result(removed);
  }
  /**
   * Creates a result containing removed and added entities.
   * @param removed
   * @param added
   */
  static removedAdded(removed, added) {
    return new _Result(removed, added);
  }
  /**
   * Creates an empty result.
   */
  static empty() {
    return _Result.EMPTY_RESULT;
  }
  /**
   * Appends remove/added items from the supplied result to the current result.
   * @param result
   */
  chain(result) {
    if (result === _Result.EMPTY_RESULT) {
      return this;
    } else if (this === _Result.EMPTY_RESULT) {
      return result;
    }
    return new _Result(
      chain(this.removed, result.removed),
      chain(this.added, result.added)
    );
  }
  /**
   * Executes given function for each added item.
   * @param fn
   *
   */
  forEachAdded(fn) {
    forEach(fn, this.added);
    return this;
  }
  /**
   * Executes given function for each removed item.
   * @param fn
   *
   */
  forEachRemoved(fn) {
    forEach(fn, this.removed);
    return this;
  }
};
let Result = _Result;
Result.EMPTY_RESULT = new _Result();

class CacheStrategy extends HasCapacity {
  constructor(capacity) {
    super(capacity);
  }
  /**
   * Removes all items from the strategy.
   *
   */
  clear() {
    while (!this.isEmpty()) {
      this.take();
    }
  }
  /**
   * Reserves place for a new item.
   * @param value
   */
  reservePlace() {
    if (this.willBeFull()) {
      const removedItem = this.take();
      if (removedItem !== NO_VALUE) {
        return Result.removed(once(removedItem));
      }
    }
    return Result.empty();
  }
}
function withDestroyable(strategy) {
  class WithDestroyable extends strategy {
    constructor(...args) {
      super(...args);
    }
    drop(value) {
      const dropped = super.drop(value);
      if (dropped) {
        value.destroy();
      }
      return dropped;
    }
    take() {
      const item = super.take();
      if (item !== NO_VALUE) {
        item.destroy();
      }
      return item;
    }
  }
  return WithDestroyable;
}

class Storage extends HasLength {
  constructor(params = EMPTY_OBJECT, parentPaths = EMPTY_ITERABLE) {
    super();
    this.params = params;
    this.parentPaths = parentPaths;
    this.destroyed = false;
    if (this.params.onCreateStorage != null)
      this.params.onCreateStorage(this);
  }
  /**
   * Calls a `destroy` implementation that will unlink given storage from all entities
   * referencing it.
   *
   */
  destroy() {
    if (this.destroyed)
      return;
    this.destroyed = true;
    forEach((path) => path.drop(this), this.parentPaths);
    if (this.params.onRemoveStorage != null)
      this.params.onRemoveStorage(this);
  }
  /**
   * Returns `true` if supplied is weak, and thus won't be stored directly.
   * @param key
   */
  isWeak(_key) {
    return false;
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(key) {
    return this.get(key) !== NO_VALUE;
  }
}

class LeafStorage extends Storage {
  constructor(storage, strategy, params, parentPaths) {
    super(params, parentPaths);
    this.storage = storage;
    this.strategy = strategy;
    this.dropStorageValue = this.storage.drop.bind(this.storage);
  }
  /**
   * Returns amount of keys (references) stored in a map.
   *
   */
  len() {
    return this.storage.len();
  }
  /**
   * Drops an item corresponding to the supplied key.
   * @param key
   *
   */
  drop(key) {
    const isWeak = this.storage.isWeak(key);
    if (!isWeak) {
      const dropped = this.strategy.drop(key);
      if (dropped && this.params.onRemoveLeaf != null)
        this.params.onRemoveLeaf(key);
    }
    const removedFromStorage = this.storage.drop(key);
    if (isWeak) {
      if (removedFromStorage !== NO_VALUE && this.params.onRemoveLeaf != null)
        this.params.onRemoveLeaf(key);
    }
    return removedFromStorage;
  }
  /**
   * Retrieves an item corresponding to the supplied key.
   * @param key
   *
   */
  get(key) {
    const isWeak = this.storage.isWeak(key);
    const extracted = this.storage.get(key);
    if (!isWeak && extracted !== NO_VALUE) {
      this.handleResult(this.strategy.read(key));
    }
    return extracted;
  }
  /**
   * Associates supplied item with the key.
   * @param key
   * @param value
   *
   */
  set(key, value) {
    const isWeak = this.storage.isWeak(key);
    if (!isWeak)
      this.handleResult(this.strategy.write(key));
    this.storage.set(key, value);
  }
  /**
   * Removes an item from the beginning of the queue.
   *
   */
  take() {
    const key = this.strategy.take();
    if (key !== NO_VALUE) {
      this.storage.drop(key);
      if (this.params.onRemoveLeaf != null)
        this.params.onRemoveLeaf(key);
    }
    return key;
  }
  /**
   * Removes all items from the storage and cache strategy.
   *
   */
  clear() {
    this.storage.clear();
    this.strategy.clear();
  }
  /**
   * Returns an iterator over the entries.
   *
   */
  entries() {
    return this.storage.entries();
  }
  /**
   * Executes callbacks for each added/removed item.
   * @param result
   */
  handleResult(result) {
    result.forEachRemoved(this.dropStorageValue);
    if (this.params.onCreateLeaf != null) {
      result.forEachAdded(this.params.onCreateLeaf);
    }
    if (this.params.onRemoveLeaf != null) {
      result.forEachRemoved(this.params.onRemoveLeaf);
    }
  }
}

class MapStorage extends Storage {
  constructor(params, rootPath) {
    super(params, rootPath);
    this.map = /* @__PURE__ */ new Map();
  }
  /**
   * Returns amount of items stored in a map.
   *
   */
  len() {
    return this.map.size;
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(key) {
    return this.map.has(key);
  }
  /**
   * Retrieves an item corresponding to the supplied key.
   * @param key
   *
   */
  get(key) {
    if (this.map.has(key)) {
      return this.map.get(key);
    } else {
      return NO_VALUE;
    }
  }
  /**
   * Drops an item corresponding to the supplied key.
   * @param key
   *
   */
  drop(key) {
    const value = this.get(key);
    this.map.delete(key);
    return value;
  }
  /**
   * Associates supplied item with the key.
   * @param key
   * @param value
   *
   */
  set(key, value) {
    this.map.set(key, value);
  }
  /**
   * Removes all items from the storage.
   *
   */
  clear() {
    this.map.clear();
  }
  /**
   * Returns an iterator over the entries.
   *
   */
  entries() {
    return map(([key, value]) => ({ key, value }), this.map.entries());
  }
}

class WeakStorage extends Storage {
  constructor(params, rootPath) {
    super(params, rootPath);
    this.weakMap = /* @__PURE__ */ new WeakMap();
    this.addedEntries = 0;
  }
  /**
   * Returns `true` if supplied is weak, and thus won't be stored directly.
   * @param key
   */
  isWeak(_key) {
    return true;
  }
  /**
   * Returns amount of items added to the storage.
   *
   */
  len() {
    return this.addedEntries;
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(key) {
    return this.weakMap.has(key);
  }
  /**
   * Retrieves an item corresponding to the supplied key.
   * @param key
   *
   */
  get(key) {
    if (this.weakMap.has(key)) {
      return this.weakMap.get(key);
    } else {
      return NO_VALUE;
    }
  }
  /**
   * Drops an item corresponding to the supplied key.
   * @param key
   *
   */
  drop(key) {
    const value = this.get(key);
    const removed = this.weakMap.delete(key);
    if (removed) {
      this.addedEntries--;
    }
    return value;
  }
  /**
   * Associates supplied item with the key.
   * @param key
   * @param value
   *
   */
  set(key, value) {
    if (!this.weakMap.has(key)) {
      this.addedEntries++;
    }
    this.weakMap.set(key, value);
  }
  /**
   * Does nothing.
   *
   */
  clear() {
  }
  /**
   * Returns an iterator over the entries.
   *
   */
  entries() {
    return empty();
  }
}

class ObjectStorage extends Storage {
  constructor(params, rootPath) {
    super(params, rootPath);
    this.length = 0;
    this.map = /* @__PURE__ */ Object.create(null);
  }
  /**
   * Returns amount of items stored in a map.
   *
   */
  len() {
    return this.length;
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param rawKey
   *
   */
  has(rawKey) {
    const objKey = this.objectKeyFromRawKey(rawKey);
    return objKey in this.map;
  }
  /**
   * Retrieves an item corresponding to the supplied key.
   * @param rawKey
   *
   */
  get(rawKey) {
    const objKey = this.objectKeyFromRawKey(rawKey);
    if (objKey in this.map) {
      return this.map[objKey];
    } else {
      return NO_VALUE;
    }
  }
  /**
   * Drops an item corresponding to the supplied key.
   * @param key
   *
   */
  drop(rawKey) {
    const objKey = this.objectKeyFromRawKey(rawKey);
    let value = NO_VALUE;
    if (objKey in this.map) {
      this.length--;
      value = this.map[objKey];
      delete this.map[objKey];
    }
    return value;
  }
  /**
   * Associates supplied item with the key.
   * @param rawKey
   * @param value
   *
   */
  set(rawKey, value) {
    const objKey = this.objectKeyFromRawKey(rawKey);
    const has = objKey in this.map;
    if (!has) {
      this.length++;
      this.map[objKey] = value;
    }
  }
  /**
   * Removes all items from the storage.
   *
   */
  clear() {
    this.length = 0;
    this.map = /* @__PURE__ */ Object.create(null);
  }
  /**
   * Returns an iterator over the entries.
   *
   */
  entries() {
    return map(
      (key) => ({
        key: this.rawKeyFromObjectKey(key),
        value: this.map[key]
      }),
      Reflect.ownKeys(this.map)
    );
  }
  /**
   * Converts provided raw key to the object key.
   * @param key
   */
  objectKeyFromRawKey(key) {
    const type = typeof key;
    switch (type) {
      case "number":
        return `0${key}`;
      case "string":
        return `1${key}`;
      case "boolean":
        return `2${key}`;
      case "bigint":
        return `3${key}`;
      case "undefined":
        return "4";
      case "symbol":
        return key;
      default:
        if (key === null)
          return "5";
        throw new TypeError(
          `Invalid value ${String(key)} with type ${type}, expected primitive`
        );
    }
  }
  /**
   * Instantiates a raw key from the provided object key.
   */
  rawKeyFromObjectKey(key) {
    switch (typeof key) {
      case "symbol":
        return key;
      case "string":
        const type = +key[0];
        switch (type) {
          case 0:
            return Number(key.slice(1));
          case 1:
            return key.slice(1);
          case 2:
            return key.slice(1) === "true";
          case 3:
            return BigInt(key.slice(1));
          case 4:
            return void 0;
          case 5:
            return null;
          default:
            throw new TypeError(`Invalid type ${type} of key ${key}`);
        }
      default:
        throw new TypeError(`Invalid key ${key}, expected string or symbol`);
    }
  }
}

class UnifiedStorage extends Storage {
  constructor(params, rootPath) {
    super(params, rootPath);
    const isMapImplemented = mapImplemented();
    const NonPrimitiveStorage = (params == null ? void 0 : params.useWeakStorage) || !isMapImplemented ? WeakStorage : MapStorage;
    const PrimitiveStorage = (params == null ? void 0 : params.useObjectStorage) || !isMapImplemented ? ObjectStorage : MapStorage;
    this.nonPrimitiveStorage = new NonPrimitiveStorage();
    this.primitiveStorage = new PrimitiveStorage();
  }
  /**
   * Returns `true` if supplied is weak, and thus won't be stored directly.
   * @param key
   */
  isWeak(key) {
    return !isPrimitiveValue(key);
  }
  /**
   * Returns amount of items stored in a map.
   *
   */
  len() {
    return this.primitiveStorage.len() + this.nonPrimitiveStorage.len();
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(key) {
    return isPrimitiveValue(key) ? this.primitiveStorage.has(key) : this.nonPrimitiveStorage.has(key);
  }
  /**
   * Retrieves an item corresponding to the supplied key.
   * @param key
   *
   */
  get(key) {
    return isPrimitiveValue(key) ? this.primitiveStorage.get(key) : this.nonPrimitiveStorage.get(key);
  }
  /**
   * Drops an item corresponding to the supplied key.
   * @param key
   *
   */
  drop(key) {
    return isPrimitiveValue(key) ? this.primitiveStorage.drop(key) : this.nonPrimitiveStorage.drop(key);
  }
  /**
   * Associates supplied item with the key.
   * @param key
   * @param value
   *
   */
  set(key, value) {
    if (isPrimitiveValue(key)) {
      this.primitiveStorage.set(key, value);
    } else {
      this.nonPrimitiveStorage.set(key, value);
    }
  }
  /**
   * Removes all items from the storage.
   *
   */
  clear() {
    this.primitiveStorage.clear();
    this.nonPrimitiveStorage.clear();
  }
  /**
   * Returns an iterator over the entries.
   *
   */
  entries() {
    return chain(
      this.primitiveStorage.entries(),
      this.nonPrimitiveStorage.entries()
    );
  }
}

const SafeMapStorage = mapImplemented() ? MapStorage : UnifiedStorage;

class RootLeafStrategy extends CacheStrategy {
  constructor(leafCapacity, strategy) {
    super(leafCapacity);
    this._leafStorageStrategy = strategy;
    this._leaves = 0;
  }
  /**
   * Records a read access of the supplied value.
   */
  read(value) {
    return this.handleInnerStrategyResult(
      this._leafStorageStrategy.read(value)
    );
  }
  /**
   * Records a write access of the supplied value.
   */
  write(value) {
    const res = !this.has(value) ? this.reservePlace() : Result.empty();
    this._leaves++;
    return res.chain(
      this.handleInnerStrategyResult(this._leafStorageStrategy.write(value))
    );
  }
  /**
   * Removes supplied value from the cache.
   */
  drop(value) {
    const dropped = this._leafStorageStrategy.drop(value);
    if (dropped) {
      this._leaves -= value.len();
    }
    return dropped;
  }
  /**
   * Returns underlying amount of tracked cache entries.
   */
  len() {
    return this._leaves;
  }
  /**
   * Returns `true` if supplied node exists in cache.
   */
  has(node) {
    return this._leafStorageStrategy.has(node);
  }
  /**
   * Peeks an item from the beginning of the queue.
   */
  peek() {
    return this._leafStorageStrategy.peek();
  }
  /**
   * Removes an item from the beginning of the queue.
   *
   */
  take() {
    const maybeStorage = this.peek();
    if (maybeStorage !== NO_VALUE) {
      const storage = maybeStorage;
      const removed = storage.take();
      if (removed !== NO_VALUE) {
        this._leaves--;
      }
      if (storage.isEmpty()) {
        this.drop(storage);
        return storage;
      }
    }
    return NO_VALUE;
  }
  /**
   * Handles a result produced by calling one of the inner strategy methods.
   */
  handleInnerStrategyResult(result) {
    forEach((removed) => this._leaves -= removed.len(), result.removed);
    return result;
  }
}

class OrderedCollection extends HasLength {
}
class IndexedOrderedCollection extends OrderedCollection {
}
class IndexedOrderedCollectionWithOrderedKeys extends IndexedOrderedCollection {
  keys() {
    return this.keysFront();
  }
}

class ListNode {
  constructor(value, root, prev = null, next = null) {
    this.value = value;
    this.root = root;
    this.next = next;
    this.prev = prev;
  }
  /**
   * Inserts supplied value before the current node.
   * @param value
   *
   */
  insertPrev(value) {
    const node = new ListNode(value, this.root, this.prev, this);
    if (this.prev !== null)
      this.prev.next = node;
    return this.prev = node;
  }
  /**
   * Inserts supplied value after the current node.
   * @param value
   *
   */
  insertNext(value) {
    const node = new ListNode(value, this.root, this, this.next);
    if (this.next !== null)
      this.next.prev = node;
    return this.next = node;
  }
  /**
   * Disconnects current node from its predecessor and successor.
   *
   */
  unlink() {
    if (this.next !== null)
      this.next.prev = this.prev;
    if (this.prev !== null)
      this.prev.next = this.next;
    this.prev = this.next = null;
  }
  /**
   * Disconnects current node from its predecessor, successor and root.
   *
   */
  destroy() {
    this.unlink();
    this.root = null;
  }
}
class LinkedList extends OrderedCollection {
  constructor(values = EMPTY_ITERABLE) {
    super();
    this.length = 0;
    this.head = null;
    this.tail = null;
    forEach((value) => this.pushBack(value), values);
  }
  /**
   * Pushes an node to the back of the list.
   * @param node
   *
   */
  pushBack(node) {
    this.length += 1;
    if (this.tail !== null) {
      this.tail = this.tail.insertNext(node);
    } else {
      this.head = this.tail = new ListNode(node, this);
    }
    return this.tail;
  }
  /**
   * Pushes an node to the front of the list.
   * @param node
   *
   */
  pushFront(node) {
    this.length += 1;
    if (this.head !== null) {
      this.head = this.head.insertPrev(node);
    } else {
      this.head = this.tail = new ListNode(node, this);
    }
    return this.head;
  }
  /**
   * Moves node to the front of the queue.
   * Returns `false` if node doesn't belong to the given list.
   * @param node
   *
   */
  moveFront(node) {
    if (!this.contains(node)) {
      return false;
    } else if (node === this.head) {
      return true;
    } else if (node === this.tail) {
      this.tail = this.tail.prev;
    }
    node.unlink();
    node.next = this.head;
    this.head.prev = node;
    this.head = node;
    return true;
  }
  /**
   * Moves node to the back of the queue.
   * Returns `false` if node doesn't belong to the given list.
   * @param node
   *
   */
  moveBack(node) {
    if (!this.contains(node)) {
      return false;
    } else if (node === this.tail) {
      return true;
    } else if (node === this.head) {
      this.head = this.head.next;
    }
    node.unlink();
    node.prev = this.tail;
    this.tail.next = node;
    this.tail = node;
    return true;
  }
  /**
   * Peeks a value from the front of the queue.
   * Returns either item or `null` if queue is empty.
   *
   */
  peekFront() {
    return this.head && this.head.value;
  }
  /**
   * Peeks a value from the back of the queue.
   * Returns either item or `null` if queue is empty.
   *
   */
  peekBack() {
    return this.tail && this.tail.value;
  }
  /**
   * Peeks an item from the front of the queue.
   * Returns either item or `null` if queue is empty.
   *
   */
  peekItemFront() {
    return this.head;
  }
  /**
   * Peeks an item from the end of the queue.
   * Returns either item or `null` if queue is empty.
   *
   */
  peekItemBack() {
    return this.tail;
  }
  /**
   * Takes a value from the back of the queue.
   * Returns either item or `null` if queue is empty.
   *
   */
  takeBack() {
    if (this.tail === null)
      return null;
    this.length--;
    const node = this.tail;
    if (node === this.head) {
      this.head = null;
    }
    if (node !== null) {
      this.tail = this.tail.prev;
      node.destroy();
      return node.value;
    }
    return null;
  }
  /**
   * Takes front node from the list.
   * Returns `null` if list has no nodes.
   *
   */
  takeFront() {
    if (this.head === null)
      return null;
    this.length--;
    const node = this.head;
    if (node === this.tail) {
      this.tail = null;
    }
    if (node !== null) {
      this.head = this.head.next;
      node.destroy();
      return node.value;
    }
    return null;
  }
  /**
   * Inserts given value after the supplied raw linked list node.
   * Returns `null` in case supplied node doesn't belong to this list.
   * @param node
   * @param value
   *
   */
  insertAfter(node, value) {
    if (!this.contains(node)) {
      return null;
    }
    this.length++;
    const inserted = node.insertNext(value);
    if (node === this.tail) {
      this.tail = inserted;
    }
    return inserted;
  }
  /**
   * Inserts given value before the supplied raw linked list node.
   * Returns `null` in case supplied node doesn't belong to this list.
   * @param node
   * @param value
   *
   */
  insertBefore(node, value) {
    if (!this.contains(node)) {
      return null;
    }
    this.length++;
    const inserted = node.insertPrev(value);
    if (node === this.head) {
      this.head = inserted;
    }
    return inserted;
  }
  /**
   * Returns `true` if supplied node belongs to the list.
   * @param node
   */
  contains(node) {
    return node.root === this;
  }
  /**
   * Removes given node from the list.
   * Returns `false` if node doesn't belong to the given list.
   * @param node
   */
  remove(node) {
    if (!this.contains(node)) {
      return false;
    }
    this.length--;
    if (node === this.head && this.head !== null) {
      this.head = this.head.next;
    }
    if (node === this.tail && this.tail !== null) {
      this.tail = this.tail.prev;
    }
    node.destroy();
    return true;
  }
  /**
   * Returns amount of items stored in the list.
   *
   */
  len() {
    return this.length;
  }
  /**
   * Returns an iterator over collection values starting from the end.
   */
  valuesBack() {
    const that = this;
    return {
      [Symbol.iterator]() {
        let tail = that.tail;
        return {
          next() {
            if (tail !== null) {
              const value = tail.value;
              tail = tail.prev;
              return { value, done: false };
            } else {
              return ITER_DONE_VALUE;
            }
          }
        };
      }
    };
  }
  /**
   * Returns an iterator over collection values starting from the beginning.
   */
  valuesFront() {
    const that = this;
    return {
      [Symbol.iterator]() {
        let head = that.head;
        return {
          next() {
            if (head !== null) {
              const value = head.value;
              head = head.next;
              return { value, done: false };
            } else {
              return ITER_DONE_VALUE;
            }
          }
        };
      }
    };
  }
}

class MultiKeyQueue extends IndexedOrderedCollectionWithOrderedKeys {
  constructor(values = EMPTY_ITERABLE) {
    super();
    this.list = new LinkedList();
    this.map = new SafeMapStorage();
    forEach((value) => this.pushBack(value), values);
  }
  /**
   * Returns amount of keys (references) stored in a map.
   *
   */
  len() {
    return this.map.len();
  }
  /**
   * Drops an item associated with the supplied key.
   * Returns dropped value in case of a successful drop or `NO_VALUE` if the value wasn't found.
   * @param key
   *
   */
  drop(key) {
    const maybeNoValueListNode = this.map.get(key);
    if (maybeNoValueListNode === NO_VALUE) {
      return NO_VALUE;
    }
    const listNode = maybeNoValueListNode;
    if (!this.remove(listNode)) {
      throw new Error("Inconsistency");
    }
    return listNode.value;
  }
  /**
   * Moves node to the front of the queue.
   * Returns `true` in case of success.
   * @param listNode
   *
   */
  moveFront(listNode) {
    return this.list.moveFront(listNode);
  }
  /**
   * Moves node to the back of the queue.
   * Returns `true` in case of success.
   * @param listNode
   *
   */
  moveBack(listNode) {
    return this.list.moveBack(listNode);
  }
  /**
   * Removes an item from the queue.
   * Returns `true` in case of success.
   * @param listNode
   *
   */
  remove(listNode) {
    const removed = this.list.remove(listNode);
    if (removed) {
      this.dissocKeys(listNode.value);
    }
    return removed;
  }
  /**
   * Returns `true` if supplied item belongs to the collection.
   */
  contains(listNode) {
    return this.list.contains(listNode);
  }
  /**
   * Drops supplied key from the map.
   * Item belonging to this key will be deleted only if it has no more references in the map.
   * Returns `true` in case of a successful removal or `false` if the value wasn't found.
   * @param key
   *
   */
  dropKey(key) {
    const maybeNoValueListNode = this.map.get(key);
    if (maybeNoValueListNode === NO_VALUE) {
      return false;
    }
    const listNode = maybeNoValueListNode;
    this.map.drop(key);
    if (!listNode.value.dropKey(key)) {
      throw new Error("Inconsistency");
    }
    if (listNode.value.isEmpty()) {
      this.list.remove(listNode);
    }
    return true;
  }
  /**
   * Inserts given value after the supplied raw linked list node.
   * @param node
   * @param key
   *
   */
  insertAfter(node, value) {
    const listNode = this.list.insertAfter(node, value);
    if (listNode == null) {
      return NO_VALUE;
    }
    this.assocKeys(value, listNode);
    return listNode;
  }
  /**
   * Inserts given value before the supplied raw linked list node.
   * @param node
   * @param key
   *
   */
  insertBefore(node, value) {
    const listNode = this.list.insertBefore(node, value);
    if (listNode == null) {
      return NO_VALUE;
    }
    this.assocKeys(value, listNode);
    return listNode;
  }
  /**
   * Returns value associated with the given key or `NO_VALUE` if the value wasn't found.
   * @param key
   *
   */
  get(key) {
    return this.map.get(key);
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(node) {
    return this.map.has(node);
  }
  /**
   * Adds a key for the supplied item to the beginning of its queue.
   * @param key
   * @param listNode
   *
   */
  addKeyFront(key, item) {
    if (this.map.has(key)) {
      throw new Error("Key already exists");
    } else if (!this.contains(item)) {
      return false;
    }
    const added = item.value.pushFront(key);
    if (added !== NO_VALUE) {
      this.map.set(key, item);
      return true;
    }
    return false;
  }
  /**
   * Adds a key for the supplied item to the end of its queue.
   * @param key
   * @param listNode
   *
   */
  addKeyBack(key, item) {
    if (this.map.has(key)) {
      throw new Error("Key already exists");
    } else if (!this.contains(item)) {
      return false;
    }
    const added = item.value.pushBack(key);
    if (added !== NO_VALUE) {
      this.map.set(key, item);
      return true;
    }
    return false;
  }
  /**
   * Takes a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  takeFront() {
    const item = this.list.takeFront();
    if (item != null) {
      this.dissocKeys(item);
    }
    return item == null ? NO_VALUE : item;
  }
  /**
   * Takes a value from the end of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  takeBack() {
    const item = this.list.takeBack();
    if (item != null) {
      this.dissocKeys(item);
    }
    return item == null ? NO_VALUE : item;
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peekFront() {
    const item = this.list.peekFront();
    return item == null ? NO_VALUE : item;
  }
  /**
   * Peeks a value from the end of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peekBack() {
    const item = this.list.peekBack();
    return item == null ? NO_VALUE : item;
  }
  /**
   * Peeks an item from the beginning of the collection.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  peekItemFront() {
    const item = this.list.peekItemFront();
    return item == null ? NO_VALUE : item;
  }
  /**
   * Peeks an item from the end of the collection.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  peekItemBack() {
    const item = this.list.peekItemBack();
    return item == null ? NO_VALUE : item;
  }
  /**
   * Peeks a key of the item from the beginning of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  peekKeyFront() {
    const first = this.list.peekFront();
    if (first != null) {
      return first.peekFront();
    } else {
      return NO_VALUE;
    }
  }
  /**
   * Peeks a key of the item from the end of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  peekKeyBack() {
    const first = this.list.peekBack();
    if (first != null) {
      return first.peekBack();
    } else {
      return NO_VALUE;
    }
  }
  /**
   * Takes a key of the item from the beginning of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  takeKeyFront() {
    const first = this.list.peekFront();
    if (first != null) {
      const key = first.takeFront();
      if (first.isEmpty()) {
        this.list.takeFront();
      }
      if (key !== NO_VALUE) {
        this.map.drop(key);
        return key;
      }
    }
    return NO_VALUE;
  }
  /**
   * Takes a key of the item from the end of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  takeKeyBack() {
    const last = this.list.peekBack();
    if (last != null) {
      const key = last.takeBack();
      if (last.isEmpty()) {
        this.list.takeBack();
      }
      if (key !== NO_VALUE) {
        this.map.drop(key);
        return key;
      }
    }
    return NO_VALUE;
  }
  /**
   * Returns an iterator over values.
   *
   */
  valuesFront() {
    return this.list.valuesFront();
  }
  /**
   * Returns an iterator over values.
   *
   */
  valuesBack() {
    return this.list.valuesBack();
  }
  /**
   * Returns an iterator over keys starting from the beginning.
   *
   */
  keysFront() {
    return flatMap((iter) => iter.keys(), this.list.valuesFront());
  }
  /**
   * Returns an iterator over keys starting from the end.
   *
   */
  keysBack() {
    return flatMap(
      (iter) => [...iter.keys()].reverse(),
      this.list.valuesBack()
    );
  }
  /**
   * Pushes an item to the end of the queue.
   * @param value
   *
   */
  pushBack(value) {
    const listNode = this.list.pushBack(value);
    this.assocKeys(value, listNode);
    return listNode;
  }
  /**
   * Pushes an item to the beginning of the queue.
   * @param value
   *
   */
  pushFront(value) {
    const listNode = this.list.pushFront(value);
    this.assocKeys(value, listNode);
    return listNode;
  }
  /**
   * Associates keys of the provided value with the supplied raw linked list node.
   * @param value
   * @param node
   *
   */
  assocKeys(value, node) {
    forEach((key) => {
      if (this.map.has(key)) {
        throw new Error("Key already exists");
      }
      this.map.set(key, node);
    }, value.keys());
  }
  /**
   * Dissociates keys of the provided value.
   * @param value
   * @param node
   *
   */
  dissocKeys(value) {
    forEach((key) => {
      if (!this.map.drop(key)) {
        throw new Error("Key doesn't exist");
      }
    }, value.keys());
  }
}

class Single extends IndexedOrderedCollectionWithOrderedKeys {
  constructor(value) {
    super();
    this.value = value;
  }
  pushFront(value) {
    if (this.isEmpty()) {
      return this.value = value;
    } else {
      return NO_VALUE;
    }
  }
  pushBack(value) {
    return this.pushFront(value);
  }
  takeKeyFront() {
    return this.takeFront();
  }
  takeKeyBack() {
    return this.takeFront();
  }
  peekKeyFront() {
    return this.peekFront();
  }
  peekKeyBack() {
    return this.peekBack();
  }
  peekItemFront() {
    return this.peekFront();
  }
  peekItemBack() {
    return this.peekBack();
  }
  addKeyFront(_key, _item) {
    return false;
  }
  addKeyBack(_key, _item) {
    return false;
  }
  dropKey(value) {
    if (this.has(value)) {
      this.takeFront();
      return true;
    } else {
      return false;
    }
  }
  get(value) {
    return this.has(value) ? this.value : NO_VALUE;
  }
  keysFront() {
    return this.valuesFront();
  }
  keysBack() {
    return this.valuesBack();
  }
  takeFront() {
    const value = this.value;
    this.value = NO_VALUE;
    return value;
  }
  takeBack() {
    return this.takeFront();
  }
  insertAfter(_item, _value) {
    return NO_VALUE;
  }
  insertBefore(_item, _value) {
    return NO_VALUE;
  }
  contains(item) {
    return equals(item, this.value);
  }
  has(value) {
    return equals(value, this.value);
  }
  drop(value) {
    return this.has(value) ? this.takeFront() : NO_VALUE;
  }
  peekFront() {
    return this.value;
  }
  peekBack() {
    return this.value;
  }
  moveFront(item) {
    return this.contains(item);
  }
  moveBack(item) {
    return this.contains(item);
  }
  remove(item) {
    return this.drop(item) !== NO_VALUE;
  }
  valuesFront() {
    return this.value === NO_VALUE ? empty() : once(this.value);
  }
  valuesBack() {
    return this.valuesFront();
  }
  len() {
    return Number(this.value !== NO_VALUE);
  }
}

class SingleKeyQueue extends IndexedOrderedCollectionWithOrderedKeys {
  constructor(values = EMPTY_ITERABLE) {
    super();
    this.inner = new MultiKeyQueue(
      map((value) => new Single(value), values)
    );
  }
  /**
   * Returns amount of keys (references) stored in a map.
   *
   */
  len() {
    return this.inner.len();
  }
  /**
   * Drops an item associated with the supplied key.
   * Returns dropped value in case of a successful drop or `NO_VALUE` if the value wasn't found.
   * @param key
   *
   */
  drop(key) {
    return this.inner.drop(key);
  }
  /**
   * Moves node to the front of the queue.
   * Returns `true` in case of success.
   * @param item
   *
   */
  moveFront(item) {
    return this.inner.moveFront(item);
  }
  /**
   * Moves node to the back of the queue.
   * Returns `true` in case of success.
   * @param item
   *
   */
  moveBack(item) {
    return this.inner.moveBack(item);
  }
  /**
   * Removes an item from the queue.
   * Returns `true` in case of success.
   */
  remove(item) {
    return this.inner.remove(item);
  }
  /**
   * Returns `true` if supplied item belongs to the collection.
   */
  contains(item) {
    return this.inner.contains(item);
  }
  /**
   * Drops supplied key from the map.
   * Item belonging to this key will be deleted only if it has no more references in the map.
   * Returns dropped value in case of a successful drop or `NO_VALUE` if the value wasn't found.
   * @param key
   *
   */
  dropKey(key) {
    return this.inner.dropKey(key);
  }
  /**
   * Inserts given value after the supplied raw linked list node.
   * @param node
   * @param key
   *
   */
  insertAfter(node, value) {
    return this.inner.insertAfter(node, new Single(value));
  }
  /**
   * Inserts given value before the supplied raw linked list node.
   * @param node
   * @param key
   *
   */
  insertBefore(node, value) {
    return this.inner.insertBefore(node, new Single(value));
  }
  /**
   * Returns value associated with the given key or `NO_VALUE` if the value wasn't found.
   * @param key
   *
   */
  get(key) {
    return this.inner.get(key);
  }
  /**
   * Returns `true` if value associated with the given key exists.
   * @param key
   *
   */
  has(key) {
    return this.inner.has(key);
  }
  /**
   * Adds a key for the supplied item to the beginning of its queue.
   * @param key
   * @param item
   *
   */
  addKeyFront(key, item) {
    return this.inner.addKeyFront(key, item);
  }
  /**
   * Adds a key for the supplied item to the end of its queue.
   * @param key
   * @param item
   *
   */
  addKeyBack(key, item) {
    return this.inner.addKeyBack(key, item);
  }
  /**
   * Takes a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  takeFront() {
    return this.inner.takeKeyFront();
  }
  /**
   * Takes a value from the end of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  takeBack() {
    return this.inner.takeKeyBack();
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peekFront() {
    return this.inner.peekKeyFront();
  }
  /**
   * Peeks a value from the end of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peekBack() {
    return this.inner.peekKeyBack();
  }
  /**
   * Peeks an item from the beginning of the collection.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  peekItemFront() {
    return this.inner.peekItemFront();
  }
  /**
   * Peeks an item from the end of the collection.
   * Returns either item or `NO_VALUE` if queue is empty.
   */
  peekItemBack() {
    return this.inner.peekItemBack();
  }
  /**
   * Peeks a key of the item from the beginning of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  peekKeyFront() {
    return this.inner.peekKeyFront();
  }
  /**
   * Peeks a key of the item from the end of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  peekKeyBack() {
    return this.inner.peekKeyBack();
  }
  /**
   * Takes a key of the item from the beginning of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  takeKeyFront() {
    return this.inner.takeKeyFront();
  }
  /**
   * Takes a key of the item from the end of the queue.
   * Returns either key or `NO_VALUE` if queue is empty.
   *
   */
  takeKeyBack() {
    return this.inner.takeKeyBack();
  }
  /**
   * Returns an iterator over values.
   *
   */
  valuesFront() {
    return this.inner.keysFront();
  }
  /**
   * Returns an iterator over values.
   *
   */
  valuesBack() {
    return this.inner.keysBack();
  }
  /**
   * Returns an iterator over keys starting from the beginning.
   *
   */
  keysFront() {
    return this.inner.keysFront();
  }
  /**
   * Returns an iterator over keys starting from the end.
   *
   */
  keysBack() {
    return this.inner.keysBack();
  }
  /**
   * Pushes an item to the end of the queue.
   * @param value
   *
   */
  pushBack(value) {
    return this.inner.pushBack(new Single(value));
  }
  /**
   * Pushes an item to the beginning of the queue.
   * @param value
   *
   */
  pushFront(value) {
    return this.inner.pushFront(new Single(value));
  }
}

class LRU extends CacheStrategy {
  constructor(capacity) {
    super(capacity);
    this.queue = new SingleKeyQueue();
  }
  /**
   * Returns amount of keys (references) stored in a map.
   *
   */
  len() {
    return this.queue.len();
  }
  /**
   * Returns `true` if given item exists in the queue.
   * @param node
   *
   */
  has(node) {
    return this.queue.has(node);
  }
  /**
   * Removes supplied item from the queue.
   * @param node
   *
   */
  drop(node) {
    return this.queue.drop(node) !== NO_VALUE;
  }
  /**
   * Records read access of the supplied item.
   * @param value
   *
   */
  read(value) {
    const item = this.queue.get(value);
    if (item === NO_VALUE) {
      throw new Error("`LRU`: cache entry doesn't exist");
    }
    const moved = this.queue.moveBack(item);
    if (!moved) {
      throw new Error(`\`LRU\`: failed to move the cache entry`);
    }
    return Result.empty();
  }
  /**
   * Records write access of the supplied item.
   * @param value
   *
   */
  write(value) {
    if (!this.has(value)) {
      const res = this.reservePlace();
      const pushed = this.queue.pushBack(value);
      if (pushed === NO_VALUE) {
        throw new Error(`\`LRU\`: failed to push a new cache entry`);
      }
      return res.chain(Result.added(once(value)));
    } else {
      return this.read(value);
    }
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peek() {
    return this.queue.peekKeyFront();
  }
  /**
   * Removes an item from the beginning of the queue.
   *
   */
  take() {
    return this.queue.takeKeyFront();
  }
}

class LFU extends CacheStrategy {
  constructor(capacity) {
    super(capacity);
    this.queue = new MultiKeyQueue();
  }
  /**
   * Returns amount of keys (references) stored in a map.
   *
   */
  len() {
    return this.queue.len();
  }
  /**
   * Records read access of the supplied item.
   * @param value
   *
   */
  write(value) {
    if (!this.has(value)) {
      const res = this.reservePlace();
      const head = this.queue.peekItemFront();
      if (head !== NO_VALUE) {
        const isFirstLevel = head.value.level === 1;
        if (isFirstLevel) {
          const addedHead = this.queue.addKeyBack(
            value,
            head
          );
          if (!addedHead) {
            throw new Error(`\`LFU\`: failed to modify cache head`);
          }
          return res.chain(Result.added(once(value)));
        }
      }
      const pushed = this.queue.pushFront(new LevelEntry(1, value));
      if (pushed === NO_VALUE) {
        throw new Error(`\`LFU\`: failed to create the cache head`);
      }
      return res.chain(Result.added(once(value)));
    } else {
      return this.read(value);
    }
  }
  /**
   * Records write access of the supplied item.
   * @param value
   *
   */
  read(value) {
    const maybeNoValueListNode = this.queue.get(value);
    if (maybeNoValueListNode === NO_VALUE) {
      throw new Error("`LFU`: cache entry doesn't exist");
    }
    const listNode = maybeNoValueListNode;
    const { next } = listNode;
    if (!this.queue.dropKey(value)) {
      throw new Error(
        `\`LFU\`: failed to drop a key in the current cache level`
      );
    }
    const newLevel = listNode.value.level + 1;
    if (next) {
      if (next.value.level === newLevel) {
        const added = this.queue.addKeyBack(value, next);
        if (!added) {
          throw new Error(
            `\`LFU\`: failed to move cache value to the next cache level`
          );
        }
      } else {
        const inserted = this.queue.insertBefore(
          next,
          new LevelEntry(newLevel, value)
        );
        if (inserted === NO_VALUE) {
          throw new Error(`\`LFU\`: failed to insert a new level of cache`);
        }
      }
    } else {
      const pushed = this.queue.pushBack(new LevelEntry(newLevel, value));
      if (pushed === NO_VALUE) {
        throw new Error(`\`LFU\`: failed to push a new cache level`);
      }
    }
    return Result.empty();
  }
  /**
   * Returns `true` if given item exists in the queue.
   * @param value
   *
   */
  has(value) {
    return this.queue.has(value);
  }
  /**
   * Removes supplied item from the queue.
   * @param value
   *
   */
  drop(value) {
    return this.queue.dropKey(value);
  }
  /**
   * Takes a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  take() {
    return this.queue.takeKeyFront();
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peek() {
    return this.queue.peekKeyFront();
  }
}
class LevelEntry extends SingleKeyQueue {
  constructor(level, value) {
    super(once(value));
    this.level = level;
  }
}

class Noop extends CacheStrategy {
  constructor(capacity) {
    super(capacity);
  }
  /**
   * Does nothing and always returns empty result.
   * @param value
   *
   */
  read(_) {
    return Result.empty();
  }
  /**
   * Does nothing and always returns empty result.
   * @param value
   *
   */
  write(_) {
    return Result.empty();
  }
  /**
   * Does nothing and always returns empty result.
   * @param node
   *
   */
  drop(_) {
    return false;
  }
  /**
   * Always returns zero.
   *
   */
  len() {
    return 0;
  }
  /**
   * Always returns `false`.
   * @param node
   *
   */
  has(_) {
    return false;
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Always returns `NO_VALUE`.
   *
   */
  peek() {
    return NO_VALUE;
  }
  /**
   * Removes an item from the beginning of the queue.
   * Always returns `NO_VALUE`.
   *
   */
  take() {
    return NO_VALUE;
  }
}

class FIFO extends CacheStrategy {
  constructor(capacity) {
    super(capacity);
    this.queue = new SingleKeyQueue();
  }
  /**
   * Returns amount of stored items.
   *
   */
  len() {
    return this.queue.len();
  }
  /**
   * Returns `true` if given item exists in the queue.
   * @param node
   *
   */
  has(node) {
    return this.queue.has(node);
  }
  /**
   * Removes supplied item from the queue.
   * @param node
   *
   */
  drop(node) {
    const has = this.queue.has(node);
    if (has) {
      this.queue.drop(node);
    }
    return has;
  }
  /**
   * Records read access of the supplied item.
   * @param value
   *
   */
  read(_value) {
    return Result.empty();
  }
  /**
   * Records write access of the supplied item.
   * @param value
   *
   */
  write(value) {
    if (!this.has(value)) {
      const res = this.reservePlace();
      const pushed = this.queue.pushBack(value);
      if (pushed === NO_VALUE) {
        throw new Error(`\`FIFO\`: failed to push a new cache entry`);
      }
      return res.chain(Result.added(once(value)));
    } else {
      const item = this.queue.get(value);
      if (item === NO_VALUE) {
        throw new Error(`\`FIFO\`: cache entry doesn't exist`);
      }
      const moved = this.queue.moveBack(item);
      if (!moved) {
        throw new Error(`\`FIFO\`: failed to move the cache entry`);
      }
      return Result.empty();
    }
  }
  /**
   * Peeks a value from the beginning of the queue.
   * Returns either item or `NO_VALUE` if queue is empty.
   *
   */
  peek() {
    return this.queue.peekKeyFront();
  }
  /**
   * Removes an item from the beginning of the queue.
   *
   */
  take() {
    return this.queue.takeKeyFront();
  }
}

const DEFAULT_MAX_ENTRIES_COUNT = 1e4;

const pickNoopIfNoLimit = (strategy, limit) => limit !== Infinity ? strategy : Noop;
const limitIsValid = (limit) => Number.isSafeInteger(limit) && limit > 0 || limit === Infinity;
const ensureLimitIsValid = (name, value) => {
  if (!limitIsValid(value)) {
    throw new Error(`\`${name}\` is invalid: ${value}`);
  }
};
const extendsCacheStrategy = (strategy) => strategy && strategy.prototype instanceof CacheStrategy;
class StorageContext {
  constructor({
    strategy = LRU,
    totalLeavesLimit = DEFAULT_MAX_ENTRIES_COUNT,
    totalStoragesLimit = Infinity,
    totalLeafStoragesLimit = Infinity,
    leavesPerStorageLimit = Infinity,
    useWeakStorage = false,
    useObjectStorage = false,
    onCreateStorage = noop,
    onRemoveStorage = noop,
    onCreateLeaf = noop,
    onRemoveLeaf = noop
  }) {
    let leafStrategyClass, storageStrategyClass;
    if (extendsCacheStrategy(strategy)) {
      leafStrategyClass = strategy;
      storageStrategyClass = strategy;
    } else {
      const config = strategy;
      leafStrategyClass = config.leafStrategyClass;
      storageStrategyClass = config.storageStrategyClass;
    }
    if (!extendsCacheStrategy(leafStrategyClass)) {
      throw new Error("Invalid `leafStrategyClass`");
    }
    if (!extendsCacheStrategy(storageStrategyClass)) {
      throw new Error("Invalid `storageStrategyClass`");
    }
    ensureLimitIsValid("totalLeavesLimit", totalLeavesLimit);
    ensureLimitIsValid("totalStoragesLimit", totalStoragesLimit);
    ensureLimitIsValid("totalLeafStoragesLimit", totalLeafStoragesLimit);
    ensureLimitIsValid("leavesPerStorageLimit", leavesPerStorageLimit);
    storageStrategyClass = withDestroyable(
      pickNoopIfNoLimit(storageStrategyClass, totalStoragesLimit)
    );
    leafStrategyClass = totalLeavesLimit === totalLeafStoragesLimit && totalLeavesLimit === leavesPerStorageLimit ? pickNoopIfNoLimit(leafStrategyClass, totalLeavesLimit) : leafStrategyClass;
    this.rootStorageStrategy = new storageStrategyClass(totalStoragesLimit);
    this.rootLeafStrategy = new RootLeafStrategy(
      totalLeavesLimit,
      new (withDestroyable(leafStrategyClass))(totalLeafStoragesLimit)
    );
    this.storageClass = useWeakStorage || useObjectStorage ? UnifiedStorage : SafeMapStorage;
    this.leafStrategyClass = leafStrategyClass;
    this.leavesPerStorageLimit = leavesPerStorageLimit;
    this.params = {
      useObjectStorage,
      useWeakStorage,
      onCreateStorage,
      onRemoveStorage,
      onCreateLeaf,
      onRemoveLeaf
    };
  }
  /**
   * Instantiates new storage with supplied parents
   * @param root
   */
  createStorage(parents = EMPTY_ITERABLE) {
    return new this.storageClass(this.params, parents);
  }
  /**
   * Instantiates new leaf storage with supplied parents
   * @param root
   */
  createLeafStorage(root = EMPTY_ITERABLE) {
    return new LeafStorage(
      this.createStorage(),
      new this.leafStrategyClass(this.leavesPerStorageLimit),
      this.params,
      root
    );
  }
}

class ResultOrPointer {
  constructor() {
    this.result = NO_VALUE;
    this.ptr = void 0;
  }
  /**
   * Produces a result.
   * @param result
   */
  static fromResult(result) {
    const that = new ResultOrPointer();
    that.result = result;
    return that;
  }
  /**
   * Produces a pointer.
   * @param ptr
   */
  static fromPointer(ptr) {
    const that = new ResultOrPointer();
    that.ptr = ptr;
    return that;
  }
}
class LinkedValue {
  constructor(value, weak) {
    this.value = value;
    this.weak = weak;
  }
}
class Last {
  constructor(length, root) {
    this.storages = Array(length);
    this.storages[0] = new LinkedValue(root, false);
    this.path = EMPTY_ARRAY;
    this.value = NO_VALUE;
  }
  /**
   * Associates storage with the provided index.
   * @param idx
   * @param storage
   */
  setStorage(idx, storage) {
    this.storages[idx] = storage;
  }
  /**
   * Retrieves storage using supplied index.
   * @param idx
   */
  storage(idx) {
    return this.storages[idx];
  }
  /**
   * Returns a value in case the path is the same as the previous one.
   * Otherwise returns an index of the greatest common node for provided and stored paths.
   * @param path
   */
  get(path) {
    const length = this.path.length;
    if (length === path.length) {
      let idx = -1;
      for (let i = 0; i < length && equals(this.path[i], path[i]); i++) {
        idx = i;
      }
      if (idx === length - 1) {
        return ResultOrPointer.fromResult(this.value);
      }
      return ResultOrPointer.fromPointer(idx === -1 ? 0 : idx);
    } else {
      return ResultOrPointer.fromPointer(0);
    }
  }
  /**
   * Updates path and stored value.
   * @param path
   * @param value
   */
  update(path, value) {
    this.path = path;
    this.value = value;
  }
  /**
   * Resets all underlying values.
   */
  reset() {
    const length = this.storages.length;
    this.storages.length = 1;
    this.storages.length = length;
    this.value = NO_VALUE;
    this.path = EMPTY_ARRAY;
  }
}
class Root {
  constructor(params, ctx) {
    this.length = params.length;
    this.ctx = ctx;
    this.root = params.length > 1 ? ctx.createStorage() : ctx.createLeafStorage();
    this.last = new Last(params.length, this.root);
    this.rootPath = new ParentPath(this.ctx.rootStorageStrategy, NO_VALUE);
    this.leafPath = new ParentPath(this.ctx.rootLeafStrategy, NO_VALUE);
    this.getOrInsertWith = params.checkLast ? this.checkLastThenGetOrInsertWith : this.getOrInsertWith;
  }
  /**
   * Retrieves cached value if already calculated, otherwise calls the supplied function with provided arguments and
   * saves the result.
   * @param path
   * @param calculate
   */
  getOrInsertWith(path, calculate) {
    if (path.length !== this.length)
      throw new Error("Invalid path length");
    const result = this.extractOrSetPath(path, calculate);
    this.last.reset();
    return result;
  }
  /**
   * - First, checks last arguments to be the same as the provided, and if so, returns cached value.
   * - Second, attempts to retrieve the cached value, otherwise calls the supplied function with provided arguments and
   * saves the result.
   * @param path
   * @param calculate
   */
  checkLastThenGetOrInsertWith(path, calculate) {
    const length = this.length;
    if (path.length !== length)
      throw new Error("Invalid path length");
    let result = NO_VALUE, ptr = 0;
    ({ result, ptr } = this.last.get(path));
    if (result !== NO_VALUE) {
      this.readCache();
    } else {
      result = this.extractOrSetPath(path, calculate, ptr);
    }
    return result;
  }
  extractOrSetPath(path, calculate, from = 0) {
    let { result, ptr } = this.extractPath(path, from);
    if (result !== NO_VALUE) {
      this.last.update(path, result);
      this.readCache();
      return result;
    }
    result = calculate.apply(null, path);
    this.setPath(path, result, ptr);
    this.last.update(path, result);
    this.writeCache();
    return result;
  }
  extractPath(path, from = 0) {
    const length = this.length;
    const node = this.last.storage(from);
    let idx = from, lessThanPath = true, res = node.value, isLastWeak = node.weak;
    do {
      this.last.setStorage(
        idx,
        new LinkedValue(res, isLastWeak)
      );
      isLastWeak = res.isWeak(path[idx]);
      res = res.get(path[idx]);
      lessThanPath = ++idx < length;
    } while (lessThanPath && res !== NO_VALUE);
    if (lessThanPath || res === NO_VALUE) {
      return ResultOrPointer.fromPointer(Math.max(0, idx - 1));
    }
    return ResultOrPointer.fromResult(res);
  }
  setPath(path, value, from = 0) {
    const length = this.length;
    let idx = from, cache = this.last.storage(from).value;
    while (idx < length - 1) {
      const current = path[idx];
      let next = cache.get(current);
      const isWeak = cache.isWeak(current);
      if (next === NO_VALUE) {
        const rootPath = isWeak ? once(this.rootPath) : double(this.rootPath, new ParentPath(cache, current));
        next = idx < length - 2 ? this.ctx.createStorage(rootPath) : this.ctx.createLeafStorage(append(rootPath, this.leafPath));
        cache.set(current, next);
      }
      cache = next;
      this.last.setStorage(++idx, new LinkedValue(cache, isWeak));
    }
    cache.set(path[length - 1], value);
  }
  readCache() {
    for (let i = this.length; i-- > 0; ) {
      const node = this.last.storage(i);
      if (!node.weak)
        this.ctx.rootStorageStrategy.read(node.value);
    }
    const leafNode = this.last.storage(this.length - 1);
    if (!leafNode.weak) {
      this.ctx.rootLeafStrategy.read(leafNode.value);
    }
  }
  writeCache() {
    for (let i = this.length; i-- > 0; ) {
      const node = this.last.storage(i);
      if (!node.weak)
        this.ctx.rootStorageStrategy.write(node.value);
    }
    const leafNode = this.last.storage(this.length - 1);
    if (!leafNode.weak) {
      this.ctx.rootLeafStrategy.write(leafNode.value);
    }
  }
}

var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __objRest = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};
const { slice } = Array.prototype;
function memoize(func, _a = EMPTY_OBJECT) {
  var _b = _a, {
    length = func.length,
    checkLast = true
  } = _b, params = __objRest(_b, [
    "length",
    "checkLast"
  ]);
  const recomputate = function() {
    ++resultFunction.recomputations;
    return func.apply(func, arguments);
  };
  let resultFunction;
  if (length === 0) {
    let value = NO_VALUE;
    resultFunction = function cachedFunction() {
      if (value !== NO_VALUE) {
        return value;
      } else {
        return value = recomputate.apply(this, arguments);
      }
    };
  } else {
    const ctx = new StorageContext(params);
    const root = new Root({ length, checkLast }, ctx);
    resultFunction = function cachedFunction() {
      const argsLength = arguments.length;
      let output;
      if (argsLength === length) {
        output = root.getOrInsertWith(
          arguments,
          recomputate
        );
      } else if (argsLength > length) {
        output = root.getOrInsertWith(
          slice.call(arguments, 0, length),
          recomputate
        );
      } else {
        output = recomputate.apply(func, arguments);
      }
      return output;
    };
  }
  try {
    void mimic(resultFunction, func);
  } finally {
    resultFunction.recomputations = 0;
    return resultFunction;
  }
}

const createMemoizedSelector = (...params) => {
  const paramsOrFunc = params.slice(-1)[0];
  let selectorFuncs = params.slice(0, -1);
  if (Array.isArray(selectorFuncs[0]))
    [selectorFuncs] = selectorFuncs;
  if (!params.length) {
    throw new Error("Must have at least one argument");
  } else if (params.length === 1) {
    if (typeof paramsOrFunc === "object")
      throw new Error("Structured selectors aren't currently supported");
  } else {
    selectorFuncs.forEach((param) => {
      if (typeof param !== "function")
        throw new Error(
          `Invalid type of param passed to memoization function: ${param} with type: ${typeof param}`
        );
    });
  }
  const haveParams = typeof paramsOrFunc === "object";
  const calculate = haveParams ? params.slice(-2)[0] : paramsOrFunc;
  const memoized = memoize(calculate, haveParams ? paramsOrFunc : void 0);
  function selector() {
    let { length } = selectorFuncs;
    const args = Array(length);
    while (length--)
      args[length] = selectorFuncs[length].apply(null, arguments);
    return memoized.apply(null, args);
  }
  selector.recomputations = () => memoized.recomputations;
  selector.dependencies = selectorFuncs;
  selector.resultFunction = calculate;
  return selector;
};

exports.CacheStrategy = CacheStrategy;
exports.DEFAULT_MAX_ENTRIES_COUNT = DEFAULT_MAX_ENTRIES_COUNT;
exports.FIFO = FIFO;
exports.LFU = LFU;
exports.LRU = LRU;
exports.createMemoizedSelector = createMemoizedSelector;
exports.default = memoize;
exports.memoize = memoize;
//# sourceMappingURL=index.js.map
